// ROBOAT: Arduino code for a GPS robotic boat
// by Riccardo Rocca, November 2011
// email: riccardo.rocca@hotmail.com
// http://users.iot.it/riccardo.rocca

#include <string.h>
#include <stdio.h>
#include <avr\io.h>
#include <avr\interrupt.h>
#include <Wire.h>
#define nop __asm__("nop\n\t");

#define DIM 20
#define DIM2 80
#define DIM3 15

struct softserial {
 byte pin_value;
 byte bit_index;
 byte byte_value;
 byte write_index;
 byte read_index;
 char array[DIM];
} GPS, Sonar;

struct tokens {
 byte ready;
 byte char_index;
 char array[DIM2];
 char *token[DIM3];
} GPS_tokens, Sonar_tokens;

const int compassAD = 0x42 >> 1; //compass address
int Heading = 0; // compass heading
int target = 0;
float distance = 0;
int delta = 0;
int deltaE = 0;

int speedL = 0; //left motor speed:  -255 0 +255
int speedR = 0; //right motor speed: -255 0 +255
int speedMAX = 150; //150
//const double speedUNI = sq(180)/(2*speedMAX);
const int speedUNI = 3; // 2 5 10 20 50 100
const int REF_DIST = 10; // 5 10
const int PRE_DIST = 5;

char key, j, move, myString[17];
unsigned long int Time;
float Lat, Lon, Speed, Course, Depth, Temp;
float dist_AD, brng_AD, dist_AB, brng_AB, dist_XT, dist_AT;

struct LatLon {
 float Lat, Lon;
} P0, P[100];

//variables to control motors in functions speed1() and speed2()
const int pwm1 = 6;
const int ina1 = 7;
const int inb1 = 4;

const int pwm2 = 5;
const int ina2 = 2;
const int inb2 = 3;

void setup()
{
// pinMode(pwm1, OUTPUT);
 pinMode(ina1, OUTPUT);
 pinMode(inb1, OUTPUT);
// pinMode(pwm2, OUTPUT);
 pinMode(ina2, OUTPUT);
 pinMode(inb2, OUTPUT);
 speedL = 0;
 speedR = 0; 

//Setup Sonar
 pinMode(8, OUTPUT);
 digitalWrite(8, HIGH);
 
//Setup Serial Ports  
 Serial.begin(9600);
 Wire.begin();
 Serial_begin(&GPS); bitClear(DDRC, DDC0);
 Serial_begin(&Sonar); bitClear(DDRC, DDC3);
 Serial_begin_interrupt();
}

void loop() {
 //if (Serial_available(&GPS)) Serial.print(Serial_read(&GPS),BYTE);
 //if (Serial_available(&GPS)) Serial_write(&CardW, Serial_read(&GPS));
 
 read_GPS();
 read_Sonar();
 read_Compass();

 while (Serial.available()) {
  key = Serial.read();
  switch (key) {
   case 'A':
    Serial.println(Time, DEC);
    Serial.println(P0.Lat, 6); 
    Serial.println(P0.Lon, 6);
    Serial.println(Course, 1);
    Serial.println(Speed, 1);
    Serial.println(Heading, DEC);
    Serial.println(Depth, 1);
    Serial.println(Temp, 1);
    break;

   case 'B':
    j = atoi(Serial_read_string(myString, sizeof(myString)));
    P[j].Lat = atof(Serial_read_string(myString, sizeof(myString)));
    P[j].Lon = atof(Serial_read_string(myString, sizeof(myString)));
    Serial.println(j, DEC);
    Serial.println(P[j].Lat, 6);
    Serial.println(P[j].Lon, 6);
    break;
    
   case 'E':
    target = Heading + atoi(Serial_read_string(myString, sizeof(myString)));
    if (target < 0) target += 360;
    if (target >= 360) target -= 360;
   case 'C':
   case 'D':
    move = key;
    break;
  }
 }

 switch (move) {
  case 'C':
   dist_AD = calc_Distance(P[j],P0);
   brng_AD = calc_Bearing(P[j],P0);
   dist_AB = calc_Distance(P[j],P[j-1]);
   brng_AB = calc_Bearing(P[j],P[j-1]);
   dist_XT = calc_CrossTrackDist(dist_AD, brng_AD, brng_AB);
   dist_AT = calc_AlongTrackDist(dist_AD, dist_XT);
   target = brng_AB - rad2deg(atan(dist_XT / REF_DIST));
   if (target < 0) target += 360;
   if (target >= 360) target -= 360;
   calc_speed();
   distance = dist_AB - dist_AT;
   if (distance < PRE_DIST && j > 0) j--;
   if (j == 0) {speedL = 0; speedR = 0;}
   break;

  case 'D':
   speedL = 0;
   speedR = 0;
   break;

  case 'E':
   calc_speed();
   break;
 }

 set_speedL();
 set_speedR();
}

char *Serial_read_string(char *s, int i) {
 char *p, *q;
 p = s;
 q = s + i - 1;
 do {
  while (!Serial.available());             // wait for input
  *p = Serial.read();       // get it
  if (*p == '\r') break;
 } while (++p < q);
 *p = '\0';                     // null terminate the string
 return s;
}

void calc_speed() {
 delta = Heading - target;
 if (delta>180) delta -= 360;
 if (delta<-180) delta += 360;
 
 if (delta<=0) {
  speedL = speedMAX;
  speedR = speedMAX+speedUNI*delta;
 }
 else {
  speedL = speedMAX-speedUNI*delta;
  speedR = speedMAX;
 }
 
 if (speedL < 0) speedL = 0;
 if (speedR < 0) speedR = 0;
}

void set_speedL() {
 if (speedL < 0) {
  analogWrite(pwm1, -speedL);
  digitalWrite(ina1, HIGH);
  digitalWrite(inb1, LOW); 
 }
 else if (speedL > 0) {
  analogWrite(pwm1, speedL);
  digitalWrite(ina1, LOW);
  digitalWrite(inb1, HIGH); 
 }
 else {
  digitalWrite(ina1, LOW);
  digitalWrite(inb1, LOW); 
 }
}

void set_speedR() {
 if (speedR < 0) {
  analogWrite(pwm2, -speedR);
  digitalWrite(ina2, LOW);
  digitalWrite(inb2, HIGH); 
 }
 else if (speedR > 0) {
  analogWrite(pwm2, speedR);
  digitalWrite(ina2, HIGH);
  digitalWrite(inb2, LOW); 
 }
 else {
  digitalWrite(ina2, LOW);
  digitalWrite(inb2, LOW); 
 }
}

void read_Compass() {
 Wire.beginTransmission(compassAD);
 Wire.send('A');
 Wire.endTransmission();
 Wire.requestFrom(compassAD, 2);
 if(2 <= Wire.available()) { 
  Heading = Wire.receive();
  Heading = Heading << 8;
  Heading += Wire.receive();
  Heading /= 10;
 }
}

void read_GPS() {
 while (Serial_available(&GPS)) {
  if (read_tokens(Serial_read(&GPS), &GPS_tokens) &&
   !strcmp(GPS_tokens.token[1],"$GPRMC") && *GPS_tokens.token[3]=='A') {
   Time = atol(GPS_tokens.token[2]);
   P0.Lat = fixDM(atof(GPS_tokens.token[4]));
   if (*GPS_tokens.token[5]=='S') P0.Lat *= -1;
   P0.Lon = fixDM(atof(GPS_tokens.token[6]));
   if (*GPS_tokens.token[7]=='W') P0.Lon *= -1;
   Speed = atof(GPS_tokens.token[8]);
   Course = atof(GPS_tokens.token[9]);
   GPS_tokens.ready = 1;
  }
 }
}

void read_Sonar() {
 while (Serial_available(&Sonar)) {
  if (read_tokens(Serial_read(&Sonar), &Sonar_tokens)) {
   if (!strcmp(Sonar_tokens.token[1],"$SDDPT")) Depth = atof(Sonar_tokens.token[2]);
   if (!strcmp(Sonar_tokens.token[1],"$SDMTW")) Temp = atof(Sonar_tokens.token[2]);
  }
 }
}

int read_tokens (char character, struct tokens *buffer) {
 char i, *p, *token;
 if (character == '\r') {
  buffer->array[buffer->char_index] = 0;
  buffer->char_index = 0;
  if (!checksum(buffer->array)) return 0;
  p = buffer->array;
  i = 0;
  while (token = strtok_r(p, ",", &p)) buffer->token[++i] = token;
  return i;
 }
 if (character == '\n') {
  buffer->char_index = 0;
  return 0;
 }
 buffer->array[buffer->char_index] = character;
 if (buffer->char_index < DIM2) (buffer->char_index)++;
 return 0;
}

int checksum(char *str) {
 int j, len, xor1, xor2;
 len = strlen(str);
 if (str[0] != '$') return 0;
 if (str[len-3] != '*') return 0;
 xor1 = 16*hex2int(str[len-2]) + hex2int(str[len-1]);
 xor2 = 0;
 for (j=1;j<len-3;j++) xor2 = xor2 ^ str[j];
 if (xor1 != xor2) return 0;
 return 1;
}

int hex2int(char a) {
 if (a>='A' && a<='F') return 10+(a-'A');
 if (a>='a' && a<='f') return 10+(a-'a');
 if (a>='0' && a<='9') return a-'0';
 return 0;
}

float fixDM(float DM) {
 int D; float F;
 F = DM /100;
 D = int (F);
 return D + (F - D)/0.6;
}

void Serial_begin(struct softserial *buffer) {
 buffer->pin_value = 0;
 buffer->bit_index = 30;
 buffer->byte_value = 0;
 buffer->write_index = 0;
 buffer->read_index = 0;
}

void Serial_begin_interrupt() {
 TCCR2A  = 2; // set CTC mode
 TCCR2B  = 2; // set prescaler = clk/8
 OCR2A   = 140; // 1/3 2400 BAUD period
 bitSet(TIMSK2, OCIE2A); // enables output compare interrupt
 sei(); // enables interrupts
}

int Serial_available(struct softserial *buffer) {
 if (buffer->write_index >= buffer->read_index) return buffer->write_index - buffer->read_index;
 else return (buffer->write_index + DIM) - buffer->read_index;
}

byte Serial_read(struct softserial *buffer) {
 byte read_byte = buffer->array[buffer->read_index];
 buffer->read_index++;
 if (buffer->read_index == DIM) buffer->read_index = 0;
 return read_byte;
}

void Serial_write(struct softserial *buffer, byte write_byte) {
 buffer->array[buffer->write_index] = write_byte;
 buffer->write_index++;
 if (buffer->write_index == DIM) buffer->write_index = 0;
}

ISR(TIMER2_COMPA_vect) {
 GPS.pin_value = bitRead(PINC, PINC0);
 Sonar.pin_value = bitRead(PINC, PINC3) ^ 1;

 read_byte(&GPS);
 read_byte(&Sonar);
}

//pin_value, bit_index, byte, write_index, read_index, 100 bytes
void read_byte(struct softserial *buffer) {
 if (buffer->pin_value + buffer->bit_index != 29) buffer->bit_index--;
 if (25>buffer->bit_index && buffer->bit_index>0 && buffer->bit_index%3==0)
  buffer->byte_value = (buffer->byte_value >> 1) + buffer->pin_value * 0x80;
 if (buffer->bit_index <= 0) {
  buffer->bit_index = 29;
  if (buffer->pin_value) {
   buffer->array[buffer->write_index] = buffer->byte_value;
   buffer->write_index++;
   if (buffer->write_index == DIM) buffer->write_index=0;
  }
 }
}

//convert degrees to radians
float deg2rad(float degrees) {
 return(degrees * PI / 180);
}

//Convert radians to degrees
float rad2deg(float radians) {
 return(radians * 180.0 / PI);
}

//Calculate distance form lat1/lon1 to lat2/lon2 using haversine formula
//Note lat1/lon1/lat2/lon2 must be in degrees
//Returns float distance in meters
float calc_Distance(struct LatLon P1, struct LatLon P2)
{
 float dlon, dlat, a;
 P1.Lat = deg2rad(P1.Lat);
 P1.Lon = deg2rad(P1.Lon);
 P2.Lat = deg2rad(P2.Lat);
 P2.Lon = deg2rad(P2.Lon);
 dlon = P2.Lon - P1.Lon;
 dlat = P2.Lat - P1.Lat;
 a = pow(sin(dlat/2),2) + cos(P1.Lat) * cos(P2.Lat) * pow(sin(dlon/2),2);
 return(6378140 * 2 * atan2(sqrt(a), sqrt(1-a)));
 //radius of the earth (6378140 meters) in feet 20925656.2
}

//Calculate bearing from lat1/lon1 to lat2/lon2
//Note lat1/lon1/lat2/lon2 must be in degrees
//Returns float bearing in degrees
float calc_Bearing(struct LatLon P1, struct LatLon P2)
{
 float bearing = 0.0;
 P1.Lat = deg2rad(P1.Lat);
 P1.Lon = deg2rad(P1.Lon);
 P2.Lat = deg2rad(P2.Lat);
 P2.Lon = deg2rad(P2.Lon);
 //determine angle
 bearing = atan2(sin(P2.Lon-P1.Lon)*cos(P2.Lat), (cos(P1.Lat)*sin(P2.Lat))-(sin(P1.Lat)*cos(P2.Lat)*cos(P2.Lon-P1.Lon)));
 //convert to degrees
 bearing = rad2deg(bearing);
 //use mod to turn -90 = 270
 bearing = fmod((bearing + 360.0), 360);
 return bearing;
}

float calc_CrossTrackDist(float distAD, float crsAD, float crsAB)
{
 distAD /= 6378140;
 crsAD = deg2rad(crsAD);
 crsAB = deg2rad(crsAB);
 return 6378140 * asin(sin(distAD)*sin(crsAD-crsAB));
}

float calc_AlongTrackDist(float distAD, float XTD)
{
 distAD /= 6378140;
 XTD /= 6378140;
 return 6378140 * asin(sqrt(pow(sin(distAD),2)-pow(sin(XTD),2))/cos(XTD));
}
