# ROBOAT: Python interface for a GPS robotic boat
# by Riccardo Rocca, November 2011
# email: riccardo.rocca@hotmail.com
# http://users.iot.it/riccardo.rocca

from Tkinter import *
from tkFileDialog import *
from serial import *
from time import *

ser = Serial('COM1', 9600, timeout=0.5)

root = Tk()
CanvasSize = 200
CanvasBorder = 10

Lat = []
Lon = []
blackLine = 0
redLine = 0
speedL = 0
speedR = 0

def readFile():
    global Lat, Lon, LonMid, LatMid, dLL, myCanvas, blackLine, redLine
    #myCanvas.delete(blackLine)
    #myCanvas.delete(redLine)
    filetypes = [('KML Files','*.kml'), ('Any File','*.*')]
    file = askopenfile(title='Choose a file',parent=console,filetypes=filetypes,mode='r')
    if file != None:
        s = ""
        for line in file.readlines():
                s += line

        ss = s[s.find('<coordinates>')+len('<coordinates>'):s.find('</coordinates>')].replace('\n','').split()
        for LatLon_string in ss:
                LatLon_array = LatLon_string.split(',')
                Lat.append(float(LatLon_array[1]))
                Lon.append(float(LatLon_array[0]))
        Lat.reverse()
        Lon.reverse()

        for i in range (len(Lat)):
            print(str(i)+", "+str(Lat[i])+", "+str(Lon[i]))

        LatMin = min(Lat)
        LatMax = max(Lat)
        LonMin = min(Lon)
        LonMax = max(Lon)
        dLat = LatMax - LatMin
        dLon = LonMax - LonMin
        LatMid = LatMin + dLat/2
        LonMid = LonMin + dLon/2
        dLL = max(dLat, dLon)

        line_coords = []
        for i in range(len(Lat)):
            line_coords.append(CanvasSize/2 + (Lon[i] - LonMid)*(CanvasSize-CanvasBorder)/dLL)
            line_coords.append(CanvasSize/2 - (Lat[i] - LatMid)*(CanvasSize-CanvasBorder)/dLL)
        blackLine = myCanvas.create_line(line_coords, fill='black')

        file.close()

def kmlUpdate():
    ser.flushInput()
    ser.flushOutput()
    ser.write('A')
    boatTime = ser.readline().rstrip('\r\n')
    boatLat = ser.readline().rstrip('\r\n')
    boatLon = ser.readline().rstrip('\r\n')
    boatCourse = ser.readline().rstrip('\r\n')
    boatSpeed = ser.readline().rstrip('\r\n')
    boatHead = ser.readline().rstrip('\r\n')
    boatDepth = ser.readline().rstrip('\r\n')
    boatTemp = ser.readline().rstrip('\r\n')

    print "%8s%11s%12s%5s%6s%4s%6s%5s" % (boatTime, boatLat, boatLon, boatSpeed, boatCourse, boatHead, boatDepth, boatTemp)
    
    kml = (
        '<?xml version="1.0" encoding="UTF-8"?>\n'
        '<kml xmlns="http://www.opengis.net/kml/2.2">\n'
        ' <Document>\n'
        '\n'
        '  <Camera>\n'
        '   <longitude>%s</longitude>\n'
        '   <latitude>%s</latitude>\n'
        '   <altitude>500</altitude>\n'
        '   <heading>0</heading>\n'
        '   <tilt>0</tilt>\n'
        '   <roll>0</roll>\n'
        '   <altitudeMode>relativeToGround</altitudeMode>\n'
        '  </Camera>\n'
        '\n'
        '  <Style id="mystyle">\n'
        '   <IconStyle>\n'
        '    <color>ffffffff</color>\n'
        '    <scale>0.8</scale>\n'
        '    <heading>%s</heading>\n'
        '    <Icon><href>roboat_symbol.png</href></Icon>\n'
        '   </IconStyle>\n'
        '   <LabelStyle><color>ffffffff</color></LabelStyle>\n'
        '  </Style>\n'
        '\n'
        '  <Placemark>\n'
        '   <name>%s</name>\n'
        '   <styleUrl>#mystyle</styleUrl>\n'
        '   <Point>\n'
        '    <coordinates>%s,%s,0</coordinates>\n'
        '   </Point>\n'
        '  </Placemark>\n'
        '\n'
        ' </Document>\n'
        '</kml>\n'
        ) %(boatLon, boatLat, boatHead, boatDepth, boatLon, boatLat)

    f = open('roboat/roboat_placemark.kml', 'w')
    f.write(kml)
    f.closed
    root.after(500, kmlUpdate)

def sendPath():
    global Lat, Lon, LonMid, LatMid, dLL, myCanvas, redLine
    line_coords = []
    for i in range(len(Lat)):
        ser.flushInput()
        ser.flushOutput()
        ser.write('B'+str(i)+'\r')
        ser.write(str(Lat[i])+'\r')
        ser.write(str(Lon[i])+'\r')
        SerialI = ser.readline().rstrip('\r\n')
        SerialLat = ser.readline().rstrip('\r\n')
        SerialLon = ser.readline().rstrip('\r\n')
        print(SerialI+", "+SerialLat+", "+SerialLon)
        line_coords.append(CanvasSize/2 + (float(SerialLon) - LonMid)*(CanvasSize-CanvasBorder)/dLL)
        line_coords.append(CanvasSize/2 - (float(SerialLat) - LatMid)*(CanvasSize-CanvasBorder)/dLL)
    print("")
    redLine = myCanvas.create_line(line_coords, fill='red', width=2)

def StartStop():
    ser.flushInput()
    ser.flushOutput()
    if (StartStopButton["text"] == "Start"):
        StartStopButton["text"] = "Stop"
        StartStopButton["background"] = "red"
        ser.write('C\r')
    else:
        StartStopButton["text"] = "Start"
        StartStopButton["background"] = "SystemButtonFace"
        ser.write('D\r')
    
def keypress(event):
    if event.keysym == 'Left':
        ser.write('E-50\r')
    if event.keysym == 'Right':
        ser.write('E+50\r')
    if event.keysym == 'Up':
        ser.write('E0\r')
    if event.keysym == 'Down':
        ser.write('D\r')

    if event.char == 'f': readFile()
    if event.char == 'p': sendPath()
    if event.char == 's': StartStop()
    
#=============================================================================
console = Frame(root, width=200, height=300)
#console.grid_propagate(0)
#console.master.title("Console")
console.grid()

myCanvas = Canvas(console, bg="white", height=CanvasSize, width=CanvasSize)
myCanvas.grid(column=0, columnspan=4, row=0)

FileOpenButton = Button(console, text='File Open', command=readFile)
FileOpenButton.grid(column=0, row=1)         

SendPathButton = Button(console, text='Send Path', command=sendPath)        
SendPathButton.grid(column=1, row=1)         

StartStopButton = Button(console, text='Start', command=StartStop)        
StartStopButton.grid(column=2, row=1)         

QuitButton = Button(console, text='Quit', command=root.destroy)       
QuitButton.grid(column=3, row=1)         

root.bind_all('<Key>', keypress)

root.after(500, kmlUpdate)
root.mainloop()

